-- Eclasses.e version 2.0
-- by Mike Nelson
-- mike-nelson-ODAAT@worldnet.att.net
-- [date]
--
-- An emulated JAVA-like class system for Euphoria.
-- Uses single inheritance with interfaces and exception handling.
-- Access control (PUBLIC, PROTECTED, PRIVATE) for members and methods.
-- Separate read and write access for members.
-- READ_ONLY write access for members.

-- NOT backward compatible with previous versions.

with trace
without warning

-- CONSTANTS:
-- Methods:
constant NO_METHOD=-1
-- Booleans:
global constant FALSE=0
global constant NO=0
global constant OFF=0
global constant NIL=0
global constant TRUE=1
global constant YES=1
global constant ON=1
-- Access rights:
global constant PUBLIC=0
global constant PROTECTED=1
global constant PRIVATE=2
global constant READ_ONLY=3
-- Program contexts:
constant MAIN_PROGRAM=0
constant CONSTRUCTOR=1
constant VALIDATOR=2
constant DESTRUCTOR=3
constant METHOD=4
-- Maximum integer:
constant MAX_INTEGER=1073741823
-- Empty lists:
global constant NONE={}

-- VARIABLES:
-- Trash can for unwanted return values:
global object VOID
-- Exception handling:
integer ExceptionsEnabled ExceptionsEnabled=TRUE
sequence PendingException PendingException=NONE
sequence CaughtException CaughtException=NONE
-- Class Database:
integer MaxClassID MaxClassID=0
sequence Class Class=NONE
sequence ClassDefined ClassDefined=NONE
sequence MaxObjectID MaxObjectID=NONE
sequence IsInterface IsInterface=NONE
sequence Superclass Superclass=NONE
sequence InterfaceList InterfaceList=NONE
sequence ConstructorID ConstructorID=NONE
sequence ValidatorID ValidatorID=NONE
sequence DestructorID DestructorID=NONE
sequence MemberList MemberList=NONE
sequence MemberReadAccess MemberReadAccess=NONE
sequence MemberWriteAccess MemberWriteAccess=NONE
sequence MethodList MethodList=NONE
sequence MethodID MethodID=NONE
sequence MethodAccess MethodAccess=NONE
sequence Password Password=rand(repeat(100000,100))+100000
-- Object database:
sequence TheObjects TheObjects=NONE
-- Program context:
sequence CallStack CallStack={{MAIN_PROGRAM,NONE,NONE,NONE}}
-- Interface implementation:
sequence RequiredMembers
sequence RequiredMethods

-- ROUTINES:
-- Utility:
global function in_range(object Param,object Min,object Max)
   return compare(Param,Min)>=0 and compare(Param,Max)<=0
end function

global function remove(sequence theSequence,integer index)
   return theSequence[1..index-1]&theSequence[index+1..length(theSequence)]
end function

global function find_duplicate(sequence theSequence)
   for i=1 to length(theSequence)-1 do
      if find(theSequence[i],theSequence[i+1..length(theSequence)]) then
	 return i
      end if
   end for
   return 0
end function

-- Types:
global type unique_sequence(object theObject)
   if sequence(theObject) then return not find_duplicate(theObject) end if
   return FALSE
end type

global type sequence_of_integers(object theObject)
   if sequence(theObject) then
      for i=1 to length(theObject) do
	 if not integer(theObject[i]) then return FALSE end if
      end for
      return TRUE
   end if
   return FALSE
end type

global type sequence_of_atoms(object theObject)
   if sequence(theObject) then
      for i=1 to length(theObject) do
	 if not atom(theObject[i]) then return FALSE end if
      end for
      return TRUE
   end if
   return FALSE
end type

global type sequence_of_sequences(object theObject)
   if sequence(theObject) then
      for i=1 to length(theObject) do
	 if not sequence(theObject[i]) then return FALSE end if
      end for
      return TRUE
   end if
   return FALSE
end type

global function sequence_of_types(integer TypeID,object theObject)
   if sequence(theObject) then
      for i=1 to length(theObject) do
	 if not call_func(TypeID,{theObject[i]}) then return FALSE end if
      end for
      return TRUE
   end if
   return FALSE
end function

global type boolean(object theObject)
   return equal(theObject,TRUE) or equal(theObject,FALSE)
end type

global type sequence_of_booleans(object theObject)
   return sequence_of_types(routine_id("boolean"),theObject)
end type

global type character(object theObject)
   return integer(theObject) and in_range(theObject,0,255)
end type

global type string(object theObject)
   return sequence_of_types(routine_id("character"),theObject)
end type

global type sequence_of_strings(object theObject)
   return sequence_of_types(routine_id("string"),theObject)
end type

global type identifier(object theObject)
   if sequence(theObject) and length(theObject)>0 then
      for i=1 to length(theObject) do
	 if not in_range(theObject[i],'A','Z')
	 and not in_range(theObject[i],'a','z') then
	    if i=1 or not find(theObject[i],"0123456789_") then
	       return FALSE
	    end if
	 end if
      end for
      return TRUE
   end if
   return FALSE
end type

global type identifier_or_none(object theObject)
   return identifier(theObject) or equal(theObject,NONE)
end type

global type sequence_of_identifiers(object theObject)
   return sequence_of_types(routine_id("identifier"),theObject)
end type

function getPassword(sequence Param)
   integer temp temp=Password[remainder(Param[1],100)+1]
   return Password[remainder(Param[2]+temp,100)+1]
end function

global type eobject(object theObject)
   if sequence_of_integers(theObject) and length(theObject)=3
   and theObject[1]>0 and theObject[2]>0 then
      return theObject[3]=getPassword(theObject[1..2])
   end if
   return FALSE
end type

global type sequence_of_eobjects(object theObject)
   return sequence_of_types(routine_id("eobject"),theObject)
end type

global type deleted_eobject(object theObject)
   if eobject(theObject) then
      return equal(TheObjects[theObject[1]][theObject[2]],NIL)
   end if
   return FALSE
end type

-- Error handling:
global procedure fatal_error(string theError)
   printf(2,"\nERROR:  "&theError&"\n",NONE)
   ? 1/0
end procedure

procedure context_check(sequence badContext,sequence message)
   sequence contextNames
   if not find(CallStack[1][1],badContext) then return end if
   contextNames={"main program","constructor","validator","destructor",
   "method"}
   fatal_error(message&" is not allowed in a "
   &contextNames[CallStack[1][1]+1]&".")
end procedure

-- Exception handling:
function formatException(sequence Param)
   sequence temp temp="\n{"
   for i=1 to length(Param) do
      temp&='"'&Param[i]&'"'
      if i<length(Param) then temp&=',' end if
   end for
   return temp&"}\n"
end function

global function exception()
   context_check({VALIDATOR},"exception()")
   return not equal(PendingException,NONE)
end function

global procedure enable_exceptions(boolean Param)
   context_check({VALIDATOR},"enable_exceptions()")
   ExceptionsEnabled=Param
   if ExceptionsEnabled=FALSE and not equal(PendingException,NONE) then
      fatal_error("Exception handling disabled while"
      &formatException(PendingException)&"exception was pending.")
   end if
end procedure

global procedure throw(sequence_of_identifiers ThrowList)
   context_check({VALIDATOR},"throw()")
   if equal(ThrowList,NONE) then
      fatal_error("Throw() list is empty.")
   end if
   if ExceptionsEnabled=FALSE then
      fatal_error(formatException(ThrowList)
      &"exception thrown while exception handling was disabled.")
   end if
   if not equal(PendingException,NONE) then
      fatal_error(formatException(ThrowList)&"exception thrown while"
      &formatException(PendingException)&"exception was pending.")
   end if
   PendingException=ThrowList
   CaughtException=NONE
end procedure

global function catch(sequence_of_identifiers CatchList)
   context_check({VALIDATOR},"catch()")
   if equal(PendingException,NONE) then return FALSE end if
   if equal(CatchList,NONE) then
      CaughtException=PendingException
      PendingException=NONE
      return TRUE
   end if
   for i=1 to length(CatchList) do
      if find(CatchList[i],PendingException) then
	 CaughtException=PendingException
	 PendingException=NONE
	 return TRUE
      end if
   end for
   return FALSE
end function

global function caught()
   context_check({VALIDATOR},"caught()")
   return CaughtException
end function

-- Default Methods:
function default_constructor(object Param)
   VOID=Param
   return TRUE
end function
constant DEF_CON=routine_id("default_constuctor")

function default_validator_destructor()
   return TRUE
end function
constant DEF_VAL_DES=routine_id("default_validator_destructor")

function default_method(object Param)
   VOID=Param
   return NIL
end function
constant DEF_METH=routine_id("default_method")

integer CALL_SUPER

global constant DEFAULT=-2
global constant SUPER=-3

-- Class Definition:
function open_definition()
   if MaxClassID>0 and ClassDefined[MaxClassID]=FALSE then
      return TRUE
   end if
   return FALSE
end function

global procedure class(identifier className,identifier_or_none superClass,
sequence_of_identifiers interfaceList)
   integer index
   sequence list
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"class()")
   if open_definition() then
      fatal_error("Class definitions may not be nested.")
   end if
   if MaxClassID=MAX_INTEGER then
      fatal_error("Too many classes defined.")
   end if
   MaxClassID+=1
   if equal(className,NONE) then
      fatal_error("Class name omitted.")
   end if
   if find(className,Class) then
      fatal_error("Class name "&className
      &" duplicates an existing class or interface name.")
   end if
   Class=append(Class,className)
   if equal(superClass,NONE) then
      Superclass&=NIL
   else
      if equal(className,superClass) then
	 fatal_error("Class "&className&" cannot be own superclass.")
      end if
      if not identifier(superClass) then
	 fatal_error("Superclass name "&superClass&" is invalid.")
      end if
      index=find(superClass,Class)
      if index=0 then
	 fatal_error("Superclass "&superClass&" does not exist.")
      end if
      if IsInterface[index]=TRUE then
	 fatal_error("Superclass "&superClass&" is an interface.")
      end if
      Superclass&=index
   end if
   list=repeat(NIL,length(interfaceList))
   for i=1 to length(interfaceList) do
      if find(interfaceList[i],interfaceList[1..i-1]) then
	 fatal_error("Interface "&interfaceList[i]&" is duplicated.")
      end if
      index=find(interfaceList[i],Class)
      if index=0 then
	 fatal_error("Interface "&interfaceList[i]&" does not exist.")
      end if
      if IsInterface[index]=FALSE then
	 fatal_error("Interface name "&interfaceList[i]
	 &" is not an inteface.")
      end if
      list[i]=index
   end for
   InterfaceList=append(InterfaceList,list)
   RequiredMembers=NONE
   RequiredMethods=NONE
   for i=1 to length(InterfaceList[MaxClassID]) do
      index=InterfaceList[MaxClassID][i]
      while index!=NIL do
	 for j=1 to length(MemberList[index]) do
	    if not find(MemberList[index][j],RequiredMembers) then
	       RequiredMembers=append(RequiredMembers,MemberList[index][j])
	    end if
	 end for
	 for j=1 to length(MethodList[index]) do
	    if not find(MethodList,RequiredMethods) then
	       RequiredMethods=append(RequiredMethods,MethodList[index][j])
	    end if
	 end for
	 index=Superclass[index]
      end while
   end for
   ClassDefined&=FALSE
   MaxObjectID&=0
   IsInterface&=FALSE
   ConstructorID&=NO_METHOD
   ValidatorID&=NO_METHOD
   DestructorID&=NO_METHOD
   MemberList=append(MemberList,NONE)
   MemberReadAccess=append(MemberReadAccess,NONE)
   MemberWriteAccess=append(MemberWriteAccess,NONE)
   MethodList=append(MethodList,NONE)
   MethodID=append(MethodID,NONE)
   MethodAccess=append(MethodAccess,NONE)
   TheObjects=append(TheObjects,NONE)
end procedure

global procedure constructor(integer ID)
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"constructor()")
   if not open_definition() then
      fatal_error("Contructor must be defined within a class definition.")
   end if
   if ConstructorID[MaxClassID]!=NO_METHOD then
      fatal_error("Constructor already defined for this class.")
   end if
   if ID=DEFAULT then ID=DEF_CON end if
   if ID=SUPER then ID=CALL_SUPER end if
   if ID<0 then fatal_error("Invalid constructor ID number.") end if
   ConstructorID[MaxClassID]=ID
end procedure

global procedure validator(integer ID)
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"validator()")
   if not open_definition() then
      fatal_error("Validator must be defined within a class definition.")
   end if
   if ValidatorID[MaxClassID]!=NO_METHOD then
      fatal_error("Validator already defined for this class.")
   end if
   if ID=DEFAULT then ID=DEF_VAL_DES end if
   if ID<0 then fatal_error("Invalid validator ID number.") end if
   ValidatorID[MaxClassID]=ID
end procedure

global procedure destructor(integer ID)
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"destructor()")
   if not open_definition() then
      fatal_error("Destructor must be defined within a class definition.")
   end if
   if DestructorID[MaxClassID]!=NO_METHOD then
      fatal_error("Destructor already defined for this class.")
   end if
   if ID=DEFAULT then ID=DEF_VAL_DES end if
   if ID<0 then fatal_error("Invalid destructor ID number.") end if
   DestructorID[MaxClassID]=ID
end procedure

global procedure member(identifier memberName,integer readAccess,
integer writeAccess)
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"member()")
   if not open_definition() then
      fatal_error("Member must be defined within a class definition.")
   end if
   if find(memberName,MemberList[MaxClassID]) then
      fatal_error("Member name "&memberName&
      " duplicates an existing member name.")
   end if
   if not in_range(readAccess,PUBLIC,PRIVATE) then
      fatal_error("Invalid read access parameter.")
   end if
   if not in_range(writeAccess,PUBLIC,READ_ONLY) then
      fatal_error("Invalid write access parameter.")
   end if
   if writeAccess<readAccess then
      fatal_error("Write access less restrictive than read access.")
   end if
   MemberList[MaxClassID]=append(MemberList[MaxClassID],memberName)
   MemberReadAccess[MaxClassID]&=readAccess
   MemberWriteAccess[MaxClassID]&=writeAccess
end procedure

global procedure method(identifier methodName,integer ID,integer access)
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"method()")
   if not open_definition() then
      fatal_error("Method must be defined within a class definition.")
   end if
   if find(methodName,MethodList[MaxClassID]) then
      fatal_error("Method name "&methodName
      &" duplicates an existing method name.")
   end if
   if ID=DEFAULT then ID=DEF_METH end if
   if ID=SUPER then ID=CALL_SUPER end if
   if ID<0 then fatal_error("Invalid method ID number.") end if
   if not in_range(access,PUBLIC,PRIVATE) then
      fatal_error("Invalid access parameter.")
   end if
   MethodList[MaxClassID]=append(MethodList[MaxClassID],methodName)
   MethodID[MaxClassID]&=ID
   MethodAccess[MaxClassID]&=access
end procedure

global procedure end_class()
   integer classIndex,mIndex
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"end_class()")
   if not open_definition() then
      fatal_error("No class definition in progress.")
   end if
   if ConstructorID[MaxClassID]=NO_METHOD then
      fatal_error("No constructor defined for this class.")
   end if
   if ValidatorID[MaxClassID]=NO_METHOD then
      fatal_error("No validator defined for this class.")
   end if
   if DestructorID[MaxClassID]=NO_METHOD then
      fatal_error("No destructor defined for this class.")
   end if
   classIndex=MaxClassID
   while classIndex!=NIL do
      for i=1 to length(MemberList[classIndex]) do
	 mIndex=find(MemberList[classIndex][i],RequiredMembers)
	 if mIndex!=NIL then
	    RequiredMembers=remove(RequiredMembers,mIndex)
	 end if
      end for
      for i=1 to length(MethodList[classIndex]) do
	 mIndex=find(MethodList[classIndex][i],RequiredMethods)
	 if mIndex!=NIL then
	    RequiredMethods=remove(RequiredMethods,mIndex)
	 end if
      end for
      classIndex=Superclass[classIndex]
   end while
   if not equal(RequiredMembers,NONE) then
      fatal_error("Required member "&RequiredMembers[1]&" not implemented.")
   end if
   if not equal(RequiredMethods,NONE) then
      fatal_error("Required method "&RequiredMethods[1]&" not implemented.")
   end if
   ClassDefined[MaxClassID]=TRUE
end procedure

global procedure interface(identifier interfaceName,
identifier_or_none superInterface,sequence_of_identifiers memberList,
sequence_of_identifiers methodList)
   integer index
   context_check({CONSTRUCTOR,VALIDATOR,DESTRUCTOR,METHOD},"interface()")
   if open_definition() then
      fatal_error("Interface definition not allowed inside a "
      &"class definition.")
   end if
   if MaxClassID=MAX_INTEGER then
      fatal_error("Too many classes defined.")
   end if
   MaxClassID+=1
   if find(interfaceName,Class) then
      fatal_error("Interface name "&interfaceName
      &" duplicates an existing class or interface name.")
   end if
   Class=append(Class,interfaceName)
   if equal(superInterface,NONE) then
      Superclass&=NIL
   else
      if equal(interfaceName,superInterface) then
	 fatal_error("Interface "&interfaceName
	 &" cannot be own superinterface.")
      end if
      index=find(superInterface,Class)
      if index=0 then
	 fatal_error("Superinterface "&superInterface&" does not exist.")
      end if
      if IsInterface[index]=FALSE then
	 fatal_error("Superinterface "&superInterface
	 &" is not an interface.")
      end if
      Superclass&=index
   end if
   if not unique_sequence(memberList) then
      fatal_error("Member name "&memberList[find_duplicate(memberList)]
      &" is duplicated.")
   end if
   MemberList=append(MemberList,memberList)
   if not unique_sequence(methodList) then
      fatal_error("Method name "&methodList[find_duplicate(methodList)]
      &" is duplicated.")
   end if
   MethodList=append(MethodList,methodList)
   ClassDefined&=TRUE
   MaxObjectID&=0
   IsInterface&=TRUE
   InterfaceList=append(InterfaceList,NONE)
   ConstructorID&=NO_METHOD
   ValidatorID&=NO_METHOD
   DestructorID&=NO_METHOD
   MemberReadAccess=append(MemberReadAccess,NONE)
   MemberWriteAccess=append(MemberWriteAccess,NONE)
   MethodID=append(MethodID,NONE)
   MethodAccess=append(MethodAccess,NONE)
   TheObjects=append(TheObjects,NONE)
end procedure

-- Class information:
global function class_name(object theObject)
   if not eobject(theObject) then return NONE end if
   return Class[theObject[1]]
end function

global function instance_of(object theObject,object theClass)
   integer classIndex,objectIndex
   classIndex=find(theClass,Class)
   if not eobject(theObject) or classIndex=NIL then return FALSE end if
   objectIndex=theObject[1]
   while objectIndex!=NIL do
      if objectIndex=classIndex then return TRUE end if
      objectIndex=Superclass[objectIndex]
   end while
   return FALSE
end function

global function implements(object theObject,object theInterface)
   integer objectIndex,interfaceIndex,flag
   sequence interfaces
   interfaceIndex=find(theInterface,Class)
   if not eobject(theObject) or interfaceIndex=NIL then return FALSE end if
   objectIndex=theObject[1]
   interfaces=NONE
   while objectIndex!=NIL do
      interfaces&=InterfaceList[objectIndex]
      objectIndex=Superclass[objectIndex]
   end while
   flag=TRUE
   while flag do
      if find(interfaceIndex,interfaces) then return TRUE end if
      flag=FALSE
      for i=1 to length(interfaces) do
	 if interfaces[i]!=NIL then
	    interfaces[i]=Superclass[interfaces[i]]
	    flag=TRUE
	 end if
      end for
   end while
   return FALSE
end function

-- CLASS OPERATIONS:
global function this()
   context_check({MAIN_PROGRAM},"this()")
   return CallStack[1][2]
end function

-- CallStack entry {context,this(),classList,methodName}
function access_rights(eobject theObject,integer theClass)
   if CallStack[1][1]=MAIN_PROGRAM then return PUBLIC end if
   if find(theClass,CallStack[1][3]) then
      if CallStack[1][1]=CONSTRUCTOR
      and equal(theObject,CallStack[1][2]) then
	 return READ_ONLY
      end if
      return PRIVATE
   end if
   if instance_of(CallStack[1][2],Class[theClass]) then
      return PROTECTED
   end if
   return PUBLIC
end function

function validate(eobject theObject)
   integer index
   sequence classList
   boolean returnValue
   if deleted_eobject(theObject) then return FALSE end if
   index=theObject[1]
   classList=NONE
   while index!=NIL do
      classList=index&classList
      CallStack=prepend(CallStack,{VALIDATOR,theObject,classList,NONE})
      returnValue=call_func(ValidatorID[index],{})
      CallStack=remove(CallStack,1)
      if returnValue=FALSE then return FALSE end if
      index=Superclass[index]
   end while
   return TRUE
end function

global function get_member(eobject theObject,identifier memberName)
   integer objectIndex,offset,memberIndex
   if deleted_eobject(theObject) then
      fatal_error("Attempted get_member() on deleted eobject.")
   end if
   objectIndex=theObject[1]
   offset=0
   while TRUE do
      memberIndex=find(memberName,MemberList[objectIndex])
      if memberIndex!=NIL then exit end if
      offset+=length(MemberList[objectIndex])
      objectIndex=Superclass[objectIndex]
      if objectIndex=NIL then
	 fatal_error("Member "&memberName&" not found.")
      end if
   end while
   if MemberReadAccess[objectIndex][memberIndex]
   >access_rights(theObject,objectIndex) then
      fatal_error("Access to member "&memberName&" denied.")
   end if
   return TheObjects[theObject[1]][theObject[2]][offset+memberIndex]
end function

global procedure set_member(eobject theObject,identifier memberName,
object newValue)
   integer objectIndex,offset,memberIndex
   object oldValue
   sequence classList
   context_check({VALIDATOR},"set_member()")
   if deleted_eobject(theObject) then
      fatal_error("Attempted set_member() on deleted eobject.")
   end if
   objectIndex=theObject[1]
   offset=0
   classList=NONE
   while TRUE do
      classList=objectIndex&classList
      memberIndex=find(memberName,MemberList[objectIndex])
      if memberIndex!=NIL then exit end if
      offset+=length(MemberList[objectIndex])
      objectIndex=Superclass[objectIndex]
      if objectIndex=NIL then
	 fatal_error("Member "&memberName&" not found.")
      end if
   end while
   if MemberWriteAccess[objectIndex][memberIndex]
   >access_rights(theObject,objectIndex) then
      fatal_error("Access to member "&memberName&" denied.")
   end if
   oldValue=TheObjects[theObject[1]][theObject[2]][offset+memberIndex]
   TheObjects[theObject[1]][theObject[2]][offset+memberIndex]=newValue
   if CallStack[1][1]=CONSTRUCTOR and equal(theObject,CallStack[1][2]) then
      return
   end if
   if validate(theObject)=TRUE then return end if
   TheObjects[theObject[1]][theObject[2]][offset+memberIndex]=oldValue
   throw({"Bad_Member_Assignment"})
end procedure

global function call_method(eobject theObject,identifier methodName,
object Param)
   integer objectIndex,methodIndex
   sequence classList
   object returnValue
   context_check({VALIDATOR},"call_method()")
   if deleted_eobject(theObject) then
      fatal_error("Attempted call_method() on deleted eobject.")
   end if
   objectIndex=theObject[1]
   classList=NONE
   while TRUE do
      classList=objectIndex&classList
      methodIndex=find(methodName,MethodList[objectIndex])
      if methodIndex!=NIL then exit end if
      objectIndex=Superclass[objectIndex]
      if objectIndex=NIL then
	 fatal_error("Method "&methodName&" not found.")
      end if
   end while
   if MethodAccess[objectIndex][methodIndex]
   >access_rights(theObject,objectIndex) then
      fatal_error("Access to method "&methodName&" denied.")
   end if
   CallStack=prepend(CallStack,{METHOD,theObject,classList,methodName})
   returnValue=call_func(MethodID[objectIndex][methodIndex],{Param})
   CallStack=remove(CallStack,1)
   return returnValue
end function

global procedure delete(eobject theObject)
   integer index
   boolean returnValue
   sequence classList
   context_check({VALIDATOR},"delete()")
   if deleted_eobject(theObject) then return end if
   index=theObject[1]
   classList=NONE
   while index!=NIL do
      classList=index&classList
      CallStack=prepend(CallStack,{DESTRUCTOR,theObject,classList,NONE})
      returnValue=call_func(DestructorID[index],{})
      CallStack=remove(CallStack,1)
      if returnValue=FALSE then
	 if validate(theObject)=FALSE then
	    fatal_error("Failed eobject deletion.")
	 end if
	 if catch({"Destructor_Error"}) then
	    throw(caught())
	 else
	    throw({"Destructor_Error"})
	 end if
	 return
      end if
      index=Superclass[index]
   end while
   TheObjects[theObject[1]][theObject[2]]=NIL
end procedure

global function new(identifier className,object Param)
   integer classIndex,objectIndex,rethrow
   sequence newObject,objectRep,classList,saveOldException,saveNewException
   boolean returnValue
   context_check({VALIDATOR},"new()")
   classIndex=find(className,Class)
   if classIndex=0 then
      fatal_error("Class name "&className&" does not exist.")
   end if
   if IsInterface[classIndex] then
      fatal_error("Class "&className&" is an interface.")
   end if
   if MaxObjectID[classIndex]=MAX_INTEGER then
      fatal_error("Too many eobjects of class "&Class[classIndex]&" created.")
   end if
   MaxObjectID[classIndex]+=1
   newObject={classIndex,MaxObjectID[classIndex]}
   newObject&=getPassword(newObject)
   objectIndex=classIndex
   objectRep=NONE
   while objectIndex!=NIL do
      objectRep&=repeat(NIL,length(MemberList[objectIndex]))
      objectIndex=Superclass[objectIndex]
   end while
   TheObjects[classIndex]=append(TheObjects[classIndex],objectRep)
   rethrow=catch({})
   saveOldException=caught()
   CallStack=prepend(CallStack,{CONSTRUCTOR,newObject,{classIndex},NONE})
   returnValue=call_func(ConstructorID[classIndex],{Param})
   CallStack=remove(CallStack,1)
   if returnValue=TRUE and validate(newObject)=TRUE then
      if rethrow then throw(saveOldException) end if
      return newObject
   end if
   if catch({"Constructor_Error"}) then
      saveNewException=caught()
   else
      saveNewException={"Constructor_Error"}
   end if
   delete(newObject)
   if catch({"Destructor_Error"}) then
      saveNewException&=caught()
   end if
   if rethrow then throw(saveOldException) end if
   throw(saveNewException)
   return newObject
end function

global function super(object Param)
   integer objectIndex,methodIndex
   object returnValue
   sequence classList
   context_check({MAIN_PROGRAM,VALIDATOR,DESTRUCTOR},"super()")
   objectIndex=Superclass[CallStack[1][3][1]]
   if objectIndex=NIL then
      fatal_error("super() is not allowed in a base class method.")
   end if
   if deleted_eobject(CallStack[1][2]) then
      fatal_error("Attempted super() on deleted eobject.")
   end if
   classList=CallStack[1][3]
   if CallStack[1][1]=CONSTRUCTOR then
      CallStack=prepend(CallStack,{CONSTRUCTOR,CallStack[1][2],
      objectIndex&classList,NONE})
      returnValue=call_func(ConstructorID[objectIndex],{Param})
      CallStack=remove(CallStack,1)
      return returnValue
   end if
   while TRUE do
      classList=objectIndex&classList
      methodIndex=find(CallStack[1][4],MethodList[objectIndex])
      if methodIndex!=NIL then exit end if
      objectIndex=Superclass[objectIndex]
      if objectIndex=NIL then
	 fatal_error("Method "&CallStack[1][4]&" not found.")
      end if
   end while
   CallStack=prepend(CallStack,{METHOD,CallStack[1][2],classList,
   CallStack[1][4]})
   returnValue=call_func(MethodID[objectIndex][methodIndex],{Param})
   CallStack=remove(CallStack,1)
   return returnValue
end function

-- Default Method:
function call_super(object Param)
   return super(Param)
end function
CALL_SUPER=routine_id("call_super")








