without warning
with trace

include wildcard.e
include eclasses.e

-- Defines the Sorted_Sequence type:
global type Sorted_Sequence(object theObject)
   if eobject(theObject) then
      return instance_of(theObject,"Sorted_Sequence")
   end if
   return FALSE
end type

-- Begin Class definiton:
class("Sorted_Sequence",NONE,NONE)

-- Class members:
member("Descending",PROTECTED,READ_ONLY)
member("Unique",PROTECTED,READ_ONLY)
member("Atom_to_Sequence",PROTECTED,READ_ONLY)
member("Ignore_Case",PROTECTED,READ_ONLY)
member("Data",PUBLIC,PRIVATE)

-- Constructor creates an empty Sorted_Sequence:
function Sorted_Sequence_new(object Param)
   sequence defaults,memberList
   defaults=append(repeat(FALSE,4),NONE)
   memberList={"Descending","Unique","Atom_to_Sequence","Ignore_Case","Data"}
   if Sorted_Sequence(Param) then
      for i=1 to 5 do
	 set_member(this(),memberList[i],get_member(Param,memberList[i]))
      end for
      return TRUE
   end if
   if atom(Param) then Param={Param} end if
   if length(Param)>4 or (length(Param)=4 and Param[3]=FALSE
   and Param[4]=TRUE) then
      return FALSE
   end if
   Param&=defaults[length(Param)+1..5]
   for i=1 to 5 do
      set_member(this(),memberList[i],Param[i])
   end for
   return TRUE
end function

constructor(routine_id("Sorted_Sequence_new"))

-- Validator function:
function Sorted_Sequence_check()
   sequence memberList
   memberList={"Descending","Unique","Atom_to_Sequence","Ignore_Case"}
   for i=1 to 4 do
      if not boolean(get_member(this(),memberList[i])) then
	 return FALSE
      end if
   end for
   if not sequence(get_member(this(),"Data")) then return FALSE end if
   return TRUE
end function

validator(routine_id("Sorted_Sequence_check"))

-- A trivial destructor:
destructor(DEFAULT)

-- Class methods:
function Sorted_Sequence_Compare(object Param)
   if get_member(this(),"Atom_to_Sequence")=TRUE then
      if atom(Param[1]) then Param[1]={Param[1]} end if
   end if
   if get_member(this(),"Ignore_Case")=TRUE then
      Param[1]=upper(Param[1])
      Param[2]=upper(Param[2])
   end if
   return compare(Param[1],Param[2])
end function

method("Compare",routine_id("Sorted_Sequence_Compare"),PROTECTED)

function Sorted_Sequence_Find(object Param)
   integer min,max,index,test
   sequence theSequence
   theSequence=get_member(this(),"Data")
   min=1
   max=length(theSequence)
   while TRUE do
      index=floor((min+max)/2)
      if max<min then return {FALSE,index} end if
      test=call_method(this(),"Compare",{Param,theSequence[index]})
      if get_member(this(),"Descending")=TRUE then test=-test end if
      if test=0 then
	 return {TRUE,index}
      elsif test=-1 then
	 max=index-1
      else
	 min=index+1
      end if
   end while
end function

method("Find",routine_id("Sorted_Sequence_Find"),PUBLIC)

function Sorted_Sequence_Add(object Param)
   sequence index,theSequence
   theSequence=get_member(this(),"Data")
   index=call_method(this(),"Find",Param)
   if get_member(this(),"Unique")=TRUE then
      if index[1]=TRUE then
	 throw({"Duplicate_Element","Sorted_Sequence_Error"})
	 return NIL
      end if
   end if
   if get_member(this(),"Atom_to_Sequence")=TRUE and atom(Param) then
      Param={Param}
   end if
   theSequence=theSequence[1..index[2]]&{Param}
   &theSequence[index[2]+1..length(theSequence)]
   set_member(this(),"Data",theSequence)
   return NIL
end function

method("Add",routine_id("Sorted_Sequence_Add"),PUBLIC)

function Sorted_Sequence_Remove(object Param)
   sequence index
   index=call_method(this(),"Find",Param)
   if index[1]=FALSE then
      throw({"Not_Found","Sorted_Sequence_Error"})
      return NIL
   end if
   set_member(this(),"Data",remove(get_member(this(),"Data"),index))
   return NIL
end function

method("Remove",routine_id("Sorted_Sequence_Remove"),PUBLIC)

function Sorted_Sequence_Get_Element(object Param)
   sequence theSequence
   theSequence=get_member(this(),"Data")
   if not integer(Param) or Param<1 or Param>length(theSequence) then
      throw({"Bad_Parameter","Sorted_Sequence_Error"})
      return NIL
   end if
   return theSequence[Param]
end function

method("Get_Element",routine_id("Sorted_Sequence_Get_Element"),PUBLIC)

end_class()

-- defines the Index_Sequence type:
global type Index_Sequence(object theObject)
    if eobject(theObject) then
      return instance_of(theObject,"Index_Sequence")
    end if
    return FALSE
end type

-- Begin class definition:
class("Index_Sequence","Sorted_Sequence",NONE)

-- Class member:
member("Index",PUBLIC,READ_ONLY)

-- Constructor creates empty Index_Sequence
function Index_Sequence_new(object Param)
   sequence defaults
   defaults={1,FALSE,FALSE,FALSE}
   if Index_Sequence(Param) then
      set_member(this(),"Index",get_member(Param,"Index"))
      return super(Param)
   end if
   if atom(Param) then Param={Param} end if
   if length(Param)>4 then return FALSE end if
   Param&=defaults[length(Param)+1..4]
   if not integer(Param[1]) or Param[1]<0 then return FALSE end if
   set_member(this(),"Index",Param[1])
   return super({Param[2],Param[3],TRUE,Param[4]})
end function

constructor(routine_id("Index_Sequence_new"))

-- Trivial validator:
validator(DEFAULT)

-- Trivial destructor:
destructor(DEFAULT)

-- Class methods:
-- Overrides Sorted_Sequence Compare
function Index_Sequence_Compare(object Param)
   integer Index
   Index=get_member(this(),"Index")
   if atom(Param[1]) then Param[1]={Param[1]} end if
   if get_member(this(),"Ignore_Case")=TRUE then
      Param[1][Index]=upper(Param[1][Index])
      Param[2][Index]=upper(Param[2][Index])
   end if
   return compare(Param[1][Index],Param[2][Index])
end function

method("Compare",routine_id("Index_Sequence_Compare"),PROTECTED)

function Index_Sequence_Add(object Param)
-- Overrides and invokes Sorted Sequence Add
   if atom(Param) then Param={Param} end if
   if length(Param)<get_member(this(),"Index") then
      throw({"Invalid_Element","Index_Sequence_Error"})
      return NIL
   end if
   return super(Param)
end function

method("Add",routine_id("Index_Sequence_Add"),PUBLIC)

function Index_Sequence_Find_Index(object Param)
   Param=append(repeat(NIL,get_member(this(),"Index")-1),Param)
   return call_method(this(),"Find","Param")
end function

method("Find_Index",routine_id("Index_Sequence_Find_Index"),PUBLIC)

function Index_Sequence_Remove_Index(object Param)
   Param=append(repeat(NIL,get_member(this(),"Index")-1),Param)
   return call_method(this(),"Remove",Param)
end function

method("Remove_Index",routine_id("Index_Sequence_Remove_Index"),PUBLIC)

end_class()

-- Key Sequence type:
global type Key_Sequence(object theObject)
   if eobject(theObject) then return
      instance_of(theObject,"Index_Sequence")
   end if
   return FALSE
end type

class("Key_Sequence","Index_Sequence",NONE)

function Key_Sequence_new(object Param)
   sequence defaults
   defaults={1,FALSE,FALSE}
   if Key_Sequence(Param) then
      return super(Param)
   end if
   if atom(Param) then Param={Param} end if
   if length(Param)>3 then return FALSE end if
   Param&=defaults[length(Param)+1..3]
   if not integer(Param[1]) or Param[1]<0 then return FALSE end if
   return super({Param[1],Param[2],TRUE,Param[3]})
end function

constructor(routine_id("Key_Sequence_new"))
validator(DEFAULT)
destructor(DEFAULT)
end_class()

